import json
import argparse
import logging
import pathlib
import tempfile
import pybase64

from typing import Any, Callable, Dict, List, Optional, Tuple

logger = logging.getLogger(__name__)


def load_image_base64(image_path: str) -> str:

    with open(image_path, "rb") as f:
        return pybase64.b64encode(f.read()).decode("utf-8")


def load_image_binary(image_path: str) -> bytes:
    with open(image_path, "rb") as f:
        return f.read()


def str2bool(val):
    """
    Resolving boolean arguments if they are not given in the standard format

    :param val: (bool or string) boolean argument type
    :type val: bool or str
    :return: (bool) the desired value {True, False}
    :rtype: bool
    """
    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        if val.lower() in ("yes", "true", "t", "y", "1"):
            return True
        if val.lower() in ("no", "false", "f", "n", "0"):
            return False
    raise argparse.ArgumentTypeError("Boolean value expected.")


def load_jsonl(file_path: pathlib.Path, source_encoding: str) -> List[Dict[str, Any]]:
    result = []
    logger.info("Loading JSON file: %s", file_path)
    with open(file_path, "r", encoding=source_encoding) as jlf:
        current_line = 0
        for l in jlf:
            logger.info("Processing line: %s", current_line)
            nxt = json.loads(l)
            result.append(nxt)
            current_line += 1
    return result


def save_jsonl(
    file_path: pathlib.Path, data: List[Dict[str, Any]], destination_encoding: str
):
    logger.info("Saving file %s", file_path)
    with open(file_path, "w", encoding=destination_encoding) as out_file:
        for i, d in enumerate(data):
            logger.info("Writing element %s", i)
            d_str = json.dumps(d, ensure_ascii=False)
            out_file.write(d_str)
            out_file.write("\n")


def line_map(
    *,
    map_func: Callable[[Dict[str, Any]], Dict[str, Any]],
    source_file: pathlib.Path,
    dest_file: pathlib.Path,
    source_encoding: str,
    dest_encoding: str,
    error_file: Optional[pathlib.Path] = None,
    error_encoding: Optional[str] = None,
) -> Tuple[int, int]:
    """
    Iterate over a JSONL file, applying map_func to each line

    :return: A tuple containing the number of lines processed and the number of lines successfully mapped.
    :rtype: Tuple[int, int]
    """
    assert source_file.exists()

    # If error_file is not specified, set up a temporary file
    def get_error_file(error_file_path: Optional[pathlib.Path]):
        if error_file_path:
            return open(error_file_path, "a", encoding=error_encoding)
        return tempfile.TemporaryFile(mode="w", encoding="utf-8-sig")

    successful_lines = 0
    error_lines = 0
    with open(source_file, "r", encoding=source_encoding) as in_file:
        with open(dest_file, "w", encoding=dest_encoding) as out_file:
            with get_error_file(error_file) as err_file:
                current_line = 0
                for nxt in in_file:
                    logger.info("Processing line: %s", current_line)
                    nxt_dict = json.loads(nxt)
                    try:
                        nxt_output = map_func(nxt_dict)
                        nxt_output_string = json.dumps(nxt_output)
                        logger.info("Writing output: %s", nxt_output_string)
                        out_file.write(nxt_output_string)
                        out_file.write("\n")
                        successful_lines += 1
                    except IOError as e:
                        logger.warning("Caught exception: %s", e)
                        err_file.write(nxt)
                        error_lines += 1
                    current_line += 1
    logger.info(
        "line_map complete (%s successes, %s failures)", successful_lines, error_lines
    )
    return successful_lines, error_lines



import os
import json
import requests
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from typing import Literal, Union, Optional, List
from langchain_core.callbacks import CallbackManagerForToolRun

DEFAULT_BING_WEB_SEARCH_ENDPOINT = "https://api.bing.microsoft.com/v7.0/search"
DEFAULT_BING_NEWS_SEARCH_ENDPOINT = "https://api.bing.microsoft.com/v7.0/news/search"
DEFAULT_BING_ENTITY_SEARCH_ENDPOINT = "https://api.bing.microsoft.com/v7.0/entities"


def tag_search_result(result: dict):
    """
    Format search results as XML.
    Args:
        result (dict): The original search result.
    Returns:
        str: The search result formatted as XML.
    """

    if result["kind"] == "web":
        title = json.dumps(result["title"], ensure_ascii=False)[1:-1]
        content = json.dumps(result["content"], ensure_ascii=False)[1:-1]
        url = result["url"]
        thumbnail_url = result["thumbnail_url"]
        source = result["source"]
        tag_result = f"<document><kind>web</kind><title>{title}</title><content>{content}</content><url>{url}</url><thumbnail_url>{thumbnail_url}</thumbnail_url><source>{source}</source></document>"
    elif result["kind"] == "news":
        title = json.dumps(result["title"], ensure_ascii=False)[1:-1]
        content = json.dumps(result["content"], ensure_ascii=False)[1:-1]
        url = result["url"]
        image = result["image"]
        tag_result = f"<document><kind>news</kind><title>{title}</title><content>{content}</content><url>{url}</url><image>{image}</image></document>"
    elif result["kind"] == "entity":
        name = json.dumps(result["name"], ensure_ascii=False)[1:-1]
        content = json.dumps(result["content"], ensure_ascii=False)[1:-1]
        url = result["url"]
        image = result["image"]
        info = json.dumps(result["info"], ensure_ascii=False)[1:-1]
        tag_result = f"<document><kind>entity</kind><name>{name}</name><content>{content}</content><url>{url}</url><image>{image}</image><info>{info}</info></document>"
    else:
        tag_result = ""

    return tag_result


class BingSearchInput(BaseModel):
    """Input for the BingSearch tool."""

    query: str = Field(description="Search query")


class BingSearch(BaseTool):
    """
    Tool that queries the Bing Search API and gets back json
    """

    name: str = "bing_search_results"
    description: str = (
        "A wrapper around Bing Search. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query. [IMPORTANT] Input(query) should be over 5 characters."
    )

    args_schema: type[BaseModel] = BingSearchInput

    api_key: Optional[str] = None
    max_results: int = 3
    locale: str = "en-US"
    include_news: bool = False
    include_entity: bool = False
    news_freshness: Optional[Literal["Day", "Week", "Month"]] = None
    topic: Literal["general", "news"] = "general"

    format_output: bool = False

    def __init__(
        self,
        api_key: Optional[str] = None,
        max_results: int = 3,
        locale: str = "en-US",
        include_news: bool = False,
        include_entity: bool = False,
        news_freshness: Optional[Literal["Day", "Week", "Month"]] = None,
        format_output: bool = False,
    ):
        super().__init__()

        if api_key is None:
            api_key = os.environ.get("BING_SUBSCRIPTION_KEY", None)

        if api_key is None:
            raise ValueError("bing_subscription_key is not set.")

        self.name = "bing_search_results"
        self.api_key = api_key
        self.max_results = max_results
        self.locale = locale
        self.include_news = include_news
        self.include_entity = include_entity
        self.news_freshness = news_freshness
        self.format_output = format_output

    def _run(
        self, query: str, run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> Union[List[dict], List[str]]:
        """Implementing BaseTool's _run(...)"""
        websearch_results = self._bing_websearch_results(
            query, max_results=self.max_results, locale=self.locale
        )

        newssearch_results = []
        entitysearch_results = []

        if self.include_news:
            newssearch_results = self._bing_newssearch_results(
                query,
                max_results=self.max_results,
                locale=self.locale,
                news_freshness=self.news_freshness,
            )
        if self.include_entity:
            entitysearch_results = self._bing_entitysearch_results(
                query, max_results=self.max_results, locale="en-US"
            )
            print(entitysearch_results)

        results = websearch_results + newssearch_results + entitysearch_results

        if self.format_output:
            return [tag_search_result(r) for r in results]
        else:
            return results

    def _bing_websearch_results(
        self, query: str, max_results: int, locale: str, **kwargs
    ) -> List[dict]:
        """
        Get Bing Web search results.
        """
        headers = {"Ocp-Apim-Subscription-Key": self.api_key}
        params = {
            "q": query,
            "count": max_results,
            "textDecorations": True,
            "textFormat": "HTML",
            "mkt": locale,
            **kwargs,
        }

        response = requests.get(
            DEFAULT_BING_WEB_SEARCH_ENDPOINT,
            headers=headers,
            params=params,
            timeout=20,  # Add a timeout of 20 seconds
        )
        response.raise_for_status()
        search_results = response.json()

        metadata_web_results = []
        if "webPages" in search_results:
            for result in search_results["webPages"]["value"]:
                metadata_result = {
                    "kind": "web",
                    "title": result["name"],
                    "content": result["snippet"],
                    "url": result["url"] if "url" in result else None,
                    "thumbnail_url": (
                        result["thumbnailUrl"] if "thumbnailUrl" in result else None
                    ),
                    "source": result["siteName"] if "siteName" in result else None,
                }
                metadata_web_results.append(metadata_result)

        return metadata_web_results

    def _bing_newssearch_results(
        self,
        query: str,
        max_results: int,
        locale: str,
        news_freshness: Optional[Literal["Day", "Week", "Month"]],
        **kwargs,
    ) -> List[dict]:
        """
        Get Bing News search results.
        """
        headers = {"Ocp-Apim-Subscription-Key": self.api_key}
        params = {
            "q": query,
            "count": max_results,
            "textDecorations": True,
            "textFormat": "HTML",
            "mkt": locale,
            **kwargs,
        }
        if news_freshness is not None:
            params["freshness"] = news_freshness

        response = requests.get(
            DEFAULT_BING_NEWS_SEARCH_ENDPOINT,
            headers=headers,
            params=params,
            timeout=20,  # Add a timeout of 10 seconds
        )
        response.raise_for_status()
        search_results = response.json()

        metadata_news_results = []
        if "value" in search_results:
            for result in search_results["value"]:
                metadata_result = {
                    "kind": "news",
                    "title": result["name"],
                    "content": result["description"],
                    "url": result["url"] if "url" in result else None,
                    "image": (
                        json.dumps(result["image"], ensure_ascii=False)
                        if "image" in result
                        else None
                    ),
                }
                metadata_news_results.append(metadata_result)

        return metadata_news_results

    def _bing_entitysearch_results(
        self, query: str, max_results: int, locale: str, **kwargs
    ) -> List[dict]:
        """
        Get Bing Entity search results.
        """
        headers = {"Ocp-Apim-Subscription-Key": self.api_key}
        params = {
            "q": query,
            "count": max_results,
            "textDecorations": True,
            "textFormat": "HTML",
            "mkt": "en-US",  # entity search usually works in en-US region.
            **kwargs,
        }
        response = requests.get(
            DEFAULT_BING_ENTITY_SEARCH_ENDPOINT,
            headers=headers,
            params=params,
            timeout=20,  # Add a timeout of 20 seconds
        )
        response.raise_for_status()
        search_results = response.json()

        metadata_entity_results = []
        if "entities" in search_results:
            for result in search_results["entities"]["value"]:
                metadata_result = {
                    "kind": "entity",
                    "name": result["name"],
                    "content": result["description"],
                    "url": result["url"] if "url" in result else None,
                    "image": (
                        json.dumps(result["image"], ensure_ascii=False)
                        if "image" in result
                        else None
                    ),
                    "info": (
                        json.dumps(result["entityPresentationInfo"], ensure_ascii=False)
                        if "entityPresentationInfo" in result
                        else None
                    ),
                }
                metadata_entity_results.append(metadata_result)

        return metadata_entity_results



import time

def format_timespan(seconds):
    hours = seconds // 3600
    minutes = (seconds - hours*3600) // 60
    remaining_seconds = seconds - hours*3600 - minutes*60
    timespan = f"{hours} hours {minutes} minutes {remaining_seconds:.4f} seconds."
    return timespan


import importlib
from typing import Any
from .qa_generator import (
    QAType,
    QADataGenerator,
    CustomQADataGenerator,
    convert_to_oai_format,
    generate_qas,
)

_module_lookup = {
    "QAType": "qa_generator.qa_types",
    "QADataGenerator": "qa_generator.QADataGenerator",
    "CustomQADataGenerator": "qa_generator.CustomQADataGenerator",
    "convert_to_oai_format": "qa_generator.convert_to_oai_format",
    "generate_qas": "qa_generator.generate_qas",
}


def __getattr__(name: str) -> Any:
    if name in _module_lookup:
        module = importlib.import_module(_module_lookup[name])
        return getattr(module, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = [
    "QAType",
    "QADataGenerator",
    "CustomQADataGenerator",
    "convert_to_oai_format",
    "generate_qas",
]



from IPython.display import Image, display
from langgraph.graph.state import CompiledStateGraph
from langchain_core.runnables.graph import CurveStyle


def visualize_graph(graph, xray=False, output_file_path=None):
    """
    Visualize a state graph.
    """
    try:
        if isinstance(graph, CompiledStateGraph):
            display(
                Image(
                    graph.get_graph(xray=xray).draw_mermaid_png(
                        background_color="white",
                        curve_style=CurveStyle.STEP,
                        output_file_path=output_file_path,
                    )
                )
            )
    except Exception as e:
        print(f"[ERROR] Visualize Graph Error: {e}")



import os
from audiomentations import (
    AddBackgroundNoise,
    OneOf,
    Compose,
    Aliasing,
    AddGaussianNoise,
    LoudnessNormalization,
    Gain,
    GainTransition,
    BandPassFilter,
    BandStopFilter,
    AddGaussianSNR,
    LowPassFilter,
    LowShelfFilter,
    HighPassFilter,
    HighShelfFilter,
    TimeStretch,
    PitchShift,
    Shift,
    AdjustDuration,
    ClippingDistortion,
    AirAbsorption,
    PeakingFilter,
)


def get_audio_augments_baseline():
    current_dir = os.path.dirname(__file__)
    augment = Compose(
        [
            OneOf(
                [
                    AddBackgroundNoise(
                        sounds_path=os.path.join(current_dir, "bg-noise.mp3"),
                        noise_rms="absolute",
                        min_absolute_rms_db=-30,
                        max_absolute_rms_db=-10,
                    ),
                    AddBackgroundNoise(
                        sounds_path=os.path.join(current_dir, "bg-noise.mp3"),
                        min_snr_db=2,
                        max_snr_db=4,
                    ),
                ],
                p=0.3,
            ),
            OneOf(
                [
                    AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=1.0),
                    AddGaussianSNR(min_snr_db=5.0, max_snr_db=40.0, p=1.0),
                    LoudnessNormalization(p=1.0),
                    Aliasing(p=1.0),
                ],
                p=0.3,
            ),
            OneOf(
                [
                    LowPassFilter(p=1.0),
                    LowShelfFilter(p=1.0),
                    HighPassFilter(p=1.0),
                    HighShelfFilter(p=1.0),
                    BandPassFilter(p=1.0),
                    BandStopFilter(p=1.0),
                    ClippingDistortion(p=0.8),
                    AirAbsorption(p=0.8),
                    PeakingFilter(p=0.8),
                ],
                p=0.6,
            ),
            OneOf(
                [
                    Gain(min_gain_db=-6.0, max_gain_db=6.0, p=1.0),
                    GainTransition(p=1.0),
                    AdjustDuration(duration_seconds=5.0, p=0.5),
                    AdjustDuration(
                        duration_seconds=5.0,
                        padding_position="start",
                        padding_mode="wrap",
                        p=0.5,
                    ),
                    AdjustDuration(
                        duration_seconds=5.0,
                        padding_position="start",
                        padding_mode="reflect",
                        p=0.5,
                    ),
                    TimeStretch(min_rate=0.8, max_rate=1.25, p=0.5),
                    PitchShift(min_semitones=-4, max_semitones=4, p=0.5),
                    Shift(p=0.5),
                ],
                p=0.3,
            ),
        ]
    )
    return augment



import os
import json
import glob
import html
import datetime
import zipfile
import shutil
import azure.cognitiveservices.speech as speechsdk
from azure_genai_utils.aoai import AOAI
from typing import Literal, Optional
from .augment import get_audio_augments_baseline
from scipy.io import wavfile
from audiomentations.core.audio_loading_utils import load_sound_file

# See https://learn.microsoft.com/en-us/azure/ai-services/speech-service/language-support?tabs=stt for supported locales
STT_LOCALE_DICT = {
    "Afrikaans (South Africa)": "af-ZA",
    "Amharic (Ethiopia)": "am-ET",
    "Arabic (United Arab Emirates)": "ar-AE",
    "Arabic (Bahrain)": "ar-BH",
    "Arabic (Algeria)": "ar-DZ",
    "Arabic (Egypt)": "ar-EG",
    "Arabic (Israel)": "ar-IL",
    "Arabic (Iraq)": "ar-IQ",
    "Arabic (Jordan)": "ar-JO",
    "Arabic (Kuwait)": "ar-KW",
    "Arabic (Lebanon)": "ar-LB",
    "Arabic (Libya)": "ar-LY",
    "Arabic (Morocco)": "ar-MA",
    "Arabic (Oman)": "ar-OM",
    "Arabic (Palestinian Authority)": "ar-PS",
    "Arabic (Qatar)": "ar-QA",
    "Arabic (Saudi Arabia)": "ar-SA",
    "Arabic (Syria)": "ar-SY",
    "Arabic (Tunisia)": "ar-TN",
    "Arabic (Yemen)": "ar-YE",
    "Assamese (India)": "as-IN",
    "Azerbaijani (Latin, Azerbaijan)": "az-AZ",
    "Bulgarian (Bulgaria)": "bg-BG",
    "Bengali (India)": "bn-IN",
    "Bosnian (Bosnia and Herzegovina)": "bs-BA",
    "Chinese": "zh-CN",
    "Chinese (Wu, Simplified)": "wuu-CN",
    "Chinese (Cantonese, Simplified)": "yue-CN",
    "Chinese (Mandarin, Simplified)": "zh-CN",
    "Chinese (Jilu Mandarin, Simplified)": "zh-CN-shandong",
    "Chinese (Southwestern Mandarin, Simplified)": "zh-CN-sichuan",
    "Chinese (Cantonese, Traditional)": "zh-HK",
    "Chinese (Taiwanese Mandarin, Traditional)": "zh-TW",
    "Catalan": "ca-ES",
    "Czech (Czechia)": "cs-CZ",
    "Welsh (United Kingdom)": "cy-GB",
    "Danish (Denmark)": "da-DK",
    "German": "de-DE",
    "German (Austria)": "de-AT",
    "German (Switzerland)": "de-CH",
    "German (Germany)": "de-DE",
    "Greek (Greece)": "el-GR",
    "English": "en-US",
    "English (Australia)": "en-AU",
    "English (Canada)": "en-CA",
    "English (United Kingdom)": "en-GB",
    "English (Ghana)": "en-GH",
    "English (Hong Kong SAR)": "en-HK",
    "English (Ireland)": "en-IE",
    "English (India)": "en-IN",
    "English (Kenya)": "en-KE",
    "English (Nigeria)": "en-NG",
    "English (New Zealand)": "en-NZ",
    "English (Philippines)": "en-PH",
    "English (Singapore)": "en-SG",
    "English (Tanzania)": "en-TZ",
    "English (United States)": "en-US",
    "English (South Africa)": "en-ZA",
    "Spanish (Argentina)": "es-AR",
    "Spanish (Bolivia)": "es-BO",
    "Spanish (Chile)": "es-CL",
    "Spanish (Colombia)": "es-CO",
    "Spanish (Costa Rica)": "es-CR",
    "Spanish (Cuba)": "es-CU",
    "Spanish (Dominican Republic)": "es-DO",
    "Spanish (Ecuador)": "es-EC",
    "Spanish (Spain)": "es-ES",
    "Spanish (Equatorial Guinea)": "es-GQ",
    "Spanish (Guatemala)": "es-GT",
    "Spanish (Honduras)": "es-HN",
    "Spanish (Mexico)": "es-MX",
    "Spanish (Nicaragua)": "es-NI",
    "Spanish (Panama)": "es-PA",
    "Spanish (Peru)": "es-PE",
    "Spanish (Puerto Rico)": "es-PR",
    "Spanish (Paraguay)": "es-PY",
    "Spanish (El Salvador)": "es-SV",
    "Spanish (United States)": "es-US",
    "Spanish (Uruguay)": "es-UY",
    "Spanish (Venezuela)": "es-VE",
    "Estonian (Estonia)": "et-EE",
    "Basque": "eu-ES",
    "Persian (Iran)": "fa-IR",
    "Finnish (Finland)": "fi-FI",
    "Filipino (Philippines)": "fil-PH",
    "French (Belgium)": "fr-BE",
    "French (Canada)": "fr-CA",
    "French (Switzerland)": "fr-CH",
    "French (France)": "fr-FR",
    "Irish (Ireland)": "ga-IE",
    "Galician": "gl-ES",
    "Gujarati (India)": "gu-IN",
    "Hebrew (Israel)": "he-IL",
    "Hindi (India)": "hi-IN",
    "Croatian (Croatia)": "hr-HR",
    "Hungarian (Hungary)": "hu-HU",
    "Armenian (Armenia)": "hy-AM",
    "Indonesian (Indonesia)": "id-ID",
    "Icelandic (Iceland)": "is-IS",
    "Italian (Switzerland)": "it-CH",
    "Italian (Italy)": "it-IT",
    "Japanese": "ja-JP",
    "Javanese (Latin, Indonesia)": "jv-ID",
    "Georgian (Georgia)": "ka-GE",
    "Kazakh (Kazakhstan)": "kk-KZ",
    "Khmer (Cambodia)": "km-KH",
    "Kannada (India)": "kn-IN",
    "Korean": "ko-KR",
    "Lao (Laos)": "lo-LA",
    "Lithuanian (Lithuania)": "lt-LT",
    "Latvian (Latvia)": "lv-LV",
    "Macedonian (North Macedonia)": "mk-MK",
    "Malayalam (India)": "ml-IN",
    "Mongolian (Mongolia)": "mn-MN",
    "Marathi (India)": "mr-IN",
    "Malay (Malaysia)": "ms-MY",
    "Maltese (Malta)": "mt-MT",
    "Burmese (Myanmar)": "my-MM",
    "Norwegian Bokmål (Norway)": "nb-NO",
    "Nepali (Nepal)": "ne-NP",
    "Dutch (Belgium)": "nl-BE",
    "Dutch (Netherlands)": "nl-NL",
    "Odia (India)": "or-IN",
    "Punjabi (India)": "pa-IN",
    "Polish (Poland)": "pl-PL",
    "Pashto (Afghanistan)": "ps-AF",
    "Portuguese (Brazil)": "pt-BR",
    "Portuguese (Portugal)": "pt-PT",
    "Romanian (Romania)": "ro-RO",
    "Russian (Russia)": "ru-RU",
    "Sinhala (Sri Lanka)": "si-LK",
    "Slovak (Slovakia)": "sk-SK",
    "Slovenian (Slovenia)": "sl-SI",
    "Albanian (Albania)": "sq-AL",
    "Serbian (Cyrillic, Bosnia and Herzegovina)": "sr-BA",
    "Serbian (Cyrillic, Serbia)": "sr-RS",
    "Sundanese (Latin, Indonesia)": "su-ID",
    "Swedish (Sweden)": "sv-SE",
    "Swahili (Kenya)": "sw-KE",
    "Swahili (Tanzania)": "sw-TZ",
    "Tamil (India)": "ta-IN",
    "Tamil (Sri Lanka)": "ta-LK",
    "Tamil (Malaysia)": "ta-MY",
    "Tamil (Singapore)": "ta-SG",
    "Telugu (India)": "te-IN",
    "Thai (Thailand)": "th-TH",
    "Tigrinya (Ethiopia)": "ti-ET",
    "Turkish (Turkey)": "tr-TR",
    "Ukrainian (Ukraine)": "uk-UA",
    "Urdu (India)": "ur-IN",
    "Urdu (Pakistan)": "ur-PK",
    "Uzbek (Latin, Uzbekistan)": "uz-UZ",
    "Vietnam": "vi-VN",
    "Xhosa (South Africa)": "xh-ZA",
    "Yiddish (World)": "yi-001",
    "Yoruba (Nigeria)": "yo-NG",
    "Zulu (South Africa)": "zu-ZA",
}


def _get_wav_file_by_speech_synthesis(
    speech_synthesizer, text, file_path, lang, tts_voice
):
    """
    Get audio file by speech synthesis
    """
    ssml = f"""<speak version='1.0' xmlns="https://www.w3.org/2001/10/synthesis" xml:lang='{lang}'>
    <voice name='{tts_voice}'>{html.escape(text)}</voice></speak>"""
    result = speech_synthesizer.speak_ssml_async(ssml).get()
    stream = speechsdk.AudioDataStream(result)

    if result.reason == speechsdk.ResultReason.SynthesizingAudioCompleted:
        print(f"[DONE] {text} | [TTS] {tts_voice} | Speech synthesized successfully.")
        stream.save_to_wav_file(file_path)
    elif result.reason == speechsdk.ResultReason.Canceled:
        cancellation_details = result.cancellation_details
        print(f"[ERROR] Error synthesizing audio: {cancellation_details.reason}")
        if cancellation_details.error_details:
            print(f"[ERROR] Details: {cancellation_details.error_details}")


class CustomSpeechToTextGenerator(AOAI):
    """
    Custom Speech to Text Generator class
    """

    def __init__(
        self,
        ai_speech_api_key: Optional[str] = None,
        ai_speech_region: Optional[str] = None,
        custom_speech_lang: str = "Korean",
        synthetic_text_file: str = "cc_support_expressions.jsonl",
        train_output_dir: str = "synthetic_data_train",
        train_output_dir_aug: str = "synthetic_data_train_aug",
        eval_output_dir: str = "synthetic_data_eval",
        **kwargs,
    ):
        """
        Initialize Custom Speech to Text Generator
        """
        super().__init__()

        self.custom_speech_lang = custom_speech_lang
        self.synthetic_text_file = synthetic_text_file

        if custom_speech_lang in STT_LOCALE_DICT:
            self.custom_speech_locale = STT_LOCALE_DICT[custom_speech_lang]
        else:
            raise ValueError(f"Unsupported language: {custom_speech_lang}")

        if ai_speech_api_key is None:
            ai_speech_api_key = os.getenv("AZURE_AI_SPEECH_API_KEY")

        if ai_speech_region is None:
            ai_speech_region = os.getenv("AZURE_AI_SPEECH_REGION")

        self.ai_speech_api_key = ai_speech_api_key
        self.ai_speech_region = ai_speech_region

        self.train_output_dir = train_output_dir
        self.train_output_dir_aug = train_output_dir_aug
        self.eval_output_dir = eval_output_dir

        try:
            # Initialize SpeechConfig
            speech_config = speechsdk.SpeechConfig(
                subscription=self.ai_speech_api_key, region=self.ai_speech_region
            )
            if not speech_config:
                raise ValueError(
                    "SpeechConfig initialization failed. Check your API key and region."
                )

            # Initialize AudioConfig
            audio_config = speechsdk.audio.AudioOutputConfig(use_default_speaker=True)
            if not audio_config:
                raise ValueError(
                    "AudioConfig initialization failed. Check your audio configuration."
                )

            # Initialize SpeechSynthesizer
            self.speech_synthesizer = speechsdk.SpeechSynthesizer(
                speech_config=speech_config, audio_config=audio_config
            )
            if not self.speech_synthesizer:
                raise ValueError("SpeechSynthesizer initialization failed.")

            print("=== Initialized CustomSpeechToTextGenerator ===")
            print(f"Train Output Directory: {self.train_output_dir}")
            print(
                f"Train Output Directory for Augmented Data: {self.train_output_dir_aug}"
            )
            print(f"Eval Output Directory for Augmented Data: {self.eval_output_dir}")

        except Exception as e:
            print(f"An error occurred during Speech SDK initialization: {e}")
            # You can choose to log the error or raise it depending on your application's needs
            raise

    @staticmethod
    def print_supported_languages():
        """
        Print supported languages
        """
        print(list(STT_LOCALE_DICT.keys()))

    def generate_synthetic_text(
        self,
        topic: str = "Call center QnA related expected spoken utterances",
        num_samples: int = 2,
        model_name: str = "gpt-4o-mini",
        max_tokens: int = 1024,
        temperature: float = 0.5,
        top_p: float = 0.9,
    ) -> Optional[str]:
        """
        Generate QnA for custom speech languages
        """

        question = f"""
        create {num_samples} lines of jsonl of the topic in {self.custom_speech_lang} and English languages. jsonl format is required. 
        use 'no' as number and '{self.custom_speech_locale}', 'en-US' keys for the languages.
        only include the lines as the result. Do not include ```jsonl, ``` and blank line in the result. 
        """

        system_message = """
        Generate plain text sentences of #topic# related text to improve the recognition of domain-specific words and phrases.
        Domain-specific words can be uncommon or made-up words, but their pronunciation must be straightforward to be recognized. 
        Use text data that's close to the expected spoken utterances. The number of utterances per line should be 1. 
        Here is examples of the expected format:
        {"no": 1, "string": "string", "string": "string"}
        {"no": 2, "string": "string", "string": "string"}
        """

        user_message = f"""
        #topic#: {topic}
        Question: {question}
        """
        response = self.client.chat.completions.create(
            model=model_name,
            messages=[
                {"role": "system", "content": system_message},
                {"role": "user", "content": user_message},
            ],
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
        )

        print("Usage Information:")

        if response.usage:
            # print(f"Cached Tokens: {response.usage.prompt_tokens_details.cached_tokens}") #only o1 models support this
            print(f"Completion Tokens: {response.usage.completion_tokens}")
            print(f"Prompt Tokens: {response.usage.prompt_tokens}")
            print(f"Total Tokens: {response.usage.total_tokens}")
            content = response.choices[0].message.content

            with open(self.synthetic_text_file, "w", encoding="utf-8") as f:
                for line in content.split("\n"):
                    if line.strip():  # Check if the line is not empty
                        f.write(line + "\n")

            return content
        else:
            print("No usage information available.")
            return None

    def save_synthetic_text(
        self,
        output_dir: str = "plain_text",
        delete_old_data: bool = True,
    ) -> None:
        """
        Save synthetic text to a file
        """
        # Check https://learn.microsoft.com/en-us/azure/ai-services/speech-service/language-support?tabs=stt for supported locale
        languages = [
            self.custom_speech_locale
        ]  # List of languages to generate audio files

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        if delete_old_data:
            for file in os.listdir(output_dir):
                os.remove(os.path.join(output_dir, file))

        with open(self.synthetic_text_file, "r", encoding="utf-8") as f:
            for line in f:
                try:
                    expression = json.loads(line)
                    for lang in languages:
                        text = expression[lang]
                        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
                        file_name = f"{lang}_{timestamp}.txt"
                        with open(
                            os.path.join(output_dir, file_name), "a", encoding="utf-8"
                        ) as plain_text:
                            plain_text.write(f"{text}\n")
                except json.JSONDecodeError as e:
                    print(f"Error decoding JSON on line: {line}")
                    print(e)

        plain_texts = glob.glob(f"{output_dir}/*.txt")

        for file_name in plain_texts:
            with open(file_name, "r", encoding="utf-8") as file:
                content = file.read()
                print(f"=== File Name: {file_name} ===")
                print(content)

    def generate_synthetic_wav(
        self,
        mode: Literal["train", "eval"] = "train",
        tts_voice_list=None,
        delete_old_data: bool = True,
    ):
        """
        Generate synthetic audio files
        """

        # Check https://learn.microsoft.com/en-us/azure/ai-services/speech-service/language-support?tabs=stt for supported locale
        language = (
            self.custom_speech_locale
        )  # List of languages to generate audio files

        # List of TTS voices to generate audio files. TTS voice can be multilingual.
        # For example, Korean text can be synthesized by 'zh-CN-XiaoxiaoMultilingualNeural' voice.
        if tts_voice_list is None:
            tts_voice_list = [
                "zh-CN-XiaoxiaoMultilingualNeural",
                "en-GB-AdaMultilingualNeural",
            ]

        output_dir = self.train_output_dir
        if mode == "eval":
            output_dir = self.eval_output_dir

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        if delete_old_data:
            for file in os.listdir(output_dir):
                os.remove(os.path.join(output_dir, file))

        for tts_voice in tts_voice_list:
            with open(self.synthetic_text_file, "r", encoding="utf-8") as f:
                for line in f:
                    try:
                        expression = json.loads(line)
                        no = expression["no"]
                        text = expression[language]
                        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
                        file_name = f"{no}_locale_{language}_speaker_{tts_voice}_{timestamp}.wav"
                        print(f"[{mode}] Generating {file_name}")
                        _get_wav_file_by_speech_synthesis(
                            self.speech_synthesizer,
                            text,
                            os.path.join(output_dir, file_name),
                            language,
                            tts_voice,
                        )
                        with open(
                            f"{output_dir}/manifest.txt", "a", encoding="utf-8"
                        ) as manifest_file:
                            manifest_file.write(f"{file_name}\t{text}\n")
                    except json.JSONDecodeError as e:
                        print(f"Error decoding JSON on line: {line}")
                        print(e)

    def augment_wav_files(
        self,
        num_augments: int = 5,
        delete_old_data: bool = True,
    ):
        """
        Augment wav files using audiomentations
        """
        orig_dir = self.train_output_dir
        aug_dir = self.train_output_dir_aug

        if not os.path.exists(aug_dir):
            os.makedirs(aug_dir)

        if delete_old_data:
            for file in os.listdir(aug_dir):
                os.remove(os.path.join(aug_dir, file))

        files = os.listdir(orig_dir)
        wav_files = [file for file in files if file.endswith(".wav")]

        # Sort wav_files by 'no' in ascending order
        wav_files.sort(key=lambda x: int(x.split("_")[0]))
        print(wav_files)

        augment = get_audio_augments_baseline()

        # Play each WAV file in the output folder
        for wav_file in wav_files:
            file_path = os.path.join(orig_dir, wav_file)
            samples, sample_rate = load_sound_file(
                file_path, sample_rate=None, mono=False
            )

            if len(samples.shape) == 2 and samples.shape[0] > samples.shape[1]:
                samples = samples.transpose()

            augmented_samples = augment(samples=samples, sample_rate=int(sample_rate))
            if len(augmented_samples.shape) == 2:
                augmented_samples = augmented_samples.transpose()

            for aug_idx in range(num_augments):
                output_file_path = os.path.join(
                    aug_dir, f"{wav_file}_aug_{aug_idx}.wav"
                )
                wavfile.write(
                    output_file_path, rate=sample_rate, data=augmented_samples
                )

        # Copy the original wav files to the augmented folder
        for f in glob.glob(f"{orig_dir}/*"):
            if os.path.isfile(f):
                shutil.copy2(f, aug_dir)

        print("Augmentation completed.")

    def package_trainset(
        self,
        use_augmented_data: bool = False,
        train_dataset_dir: str = "train_dataset",
        delete_old_data: bool = True,
    ):
        """
        Package synthetic data into a zip file
        """
        output_dir = self.train_output_dir
        if use_augmented_data:
            output_dir = self.train_output_dir_aug

        if not os.path.exists(train_dataset_dir):
            os.makedirs(train_dataset_dir)

        if delete_old_data:
            for file in os.listdir(train_dataset_dir):
                os.remove(os.path.join(train_dataset_dir, file))

        files = os.listdir(output_dir)

        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        zip_filename = f"train_{self.custom_speech_locale}_{timestamp}.zip"
        with zipfile.ZipFile(zip_filename, "w") as zipf:
            for file in files:
                zipf.write(os.path.join(output_dir, file), file)

        print(f"Created zip file: {zip_filename}")
        shutil.move(zip_filename, os.path.join(train_dataset_dir, zip_filename))
        print(f"Moved zip file to: {os.path.join(train_dataset_dir, zip_filename)}")
        train_dataset_path = {os.path.join(train_dataset_dir, zip_filename)}

        return train_dataset_path

    def package_evalset(
        self,
        eval_dataset_dir: str = "eval_dataset",
        delete_old_data: bool = True,
    ):
        """
        Package synthetic data into a zip file
        """
        output_dir = self.eval_output_dir

        if not os.path.exists(eval_dataset_dir):
            os.makedirs(eval_dataset_dir)

        if delete_old_data:
            for file in os.listdir(eval_dataset_dir):
                os.remove(os.path.join(eval_dataset_dir, file))

        files = os.listdir(output_dir)

        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        zip_filename = f"eval_{self.custom_speech_locale}_{timestamp}.zip"
        with zipfile.ZipFile(zip_filename, "w") as zipf:
            for file in files:
                zipf.write(os.path.join(output_dir, file), file)

        print(f"Created zip file: {zip_filename}")
        shutil.move(zip_filename, os.path.join(eval_dataset_dir, zip_filename))
        print(f"Moved zip file to: {os.path.join(eval_dataset_dir, zip_filename)}")
        eval_dataset_path = {os.path.join(eval_dataset_dir, zip_filename)}

        return eval_dataset_path



from langchain_core.output_parsers import StrOutputParser
from langchain_community.vectorstores import FAISS
from langchain_openai import AzureChatOpenAI, AzureOpenAIEmbeddings

from abc import ABC, abstractmethod
from operator import itemgetter
from langchain import hub


class RetrievalChain(ABC):
    """
    Base class for retrieval chains.

    :param source_uri: Source URI
    :param k: Number of results to retrieve
    :param embedding_name: Name of the embedding
    :param model_name: Name of the model
    """

    def __init__(self, **kwargs):
        self.source_uri = kwargs.get("source_uri", None)
        self.k = kwargs.get("k", 10)
        self.embedding_name = kwargs.get("embedding_name", "text-embedding-3-large")
        self.model_name = kwargs.get("model_name", "gpt-4o-mini")

    @abstractmethod
    def load_documents(self, source_uris):
        """Load documents from source URIs"""
        pass

    @abstractmethod
    def create_text_splitter(self):
        """Create a text splitter"""
        pass

    def split_documents(self, docs, text_splitter):
        """Split documents using a text splitter"""
        return text_splitter.split_documents(docs)

    def create_embedding(self):
        """Create an embedding"""
        return AzureOpenAIEmbeddings(model=self.embedding_name)

    def create_vectorstore(self, split_docs):
        """Create a vectorstore"""
        return FAISS.from_documents(
            documents=split_docs, embedding=self.create_embedding()
        )

    def create_retriever(self, vectorstore):
        """Create a retriever"""
        dense_retriever = vectorstore.as_retriever(
            search_type="similarity", search_kwargs={"k": self.k}
        )
        return dense_retriever

    def create_model(self):
        """Create a model"""
        return AzureChatOpenAI(model_name=self.model_name, temperature=0)

    def create_prompt(self):
        """Create RAG prompt"""
        return hub.pull("daekeun-ml/rag-history-baseline")

    @staticmethod
    def format_docs(docs):
        return "\n".join(docs)

    def create_chain(self):
        """Create a retrieval chain"""
        docs = self.load_documents(self.source_uri)
        text_splitter = self.create_text_splitter()
        split_docs = self.split_documents(docs, text_splitter)
        self.vectorstore = self.create_vectorstore(split_docs)
        self.retriever = self.create_retriever(self.vectorstore)
        model = self.create_model()
        prompt = self.create_prompt()
        self.chain = (
            {
                "chat_history": itemgetter("chat_history"),
                "question": itemgetter("question"),
                "context": itemgetter("context"),
            }
            | prompt
            | model
            | StrOutputParser()
        )
        return self



import os
from openai import AzureOpenAI
from dotenv import load_dotenv
from typing import Optional


class AOAI:
    """
    Helper class for Azure OpenAI
    """

    aoai_api_endpoint: Optional[str] = None
    aoai_api_key: Optional[str] = None
    aoai_api_version: Optional[str] = None
    aoai_deployment_name: Optional[str] = None
    client: AzureOpenAI

    def __init__(
        self,
        aoai_api_endpoint: Optional[str] = None,
        aoai_api_key: Optional[str] = None,
        aoai_api_version: Optional[str] = None,
        aoai_deployment_name: Optional[str] = None,
    ):
        """
        Initialize Azure OpenAI client. You can set the environment variables in .env file.
        If you do not have .env file, you can pass the values as arguments.
        Args:
            aoai_api_endpoint: Azure OpenAI endpoint
            aoai_api_key: Azure OpenAI API key
            aoai_api_version: Azure OpenAI API version
            aoai_deployment_name: Azure OpenAI deployment name
        """

        if aoai_api_endpoint is None:
            aoai_api_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")

        if aoai_api_key is None:
            aoai_api_key = os.getenv("AZURE_OPENAI_API_KEY")

        if aoai_api_version is None:
            aoai_api_version = os.getenv("AZURE_OPENAI_API_VERSION") or os.getenv(
                "OPENAI_API_VERSION"
            )

        if aoai_deployment_name is None:
            aoai_deployment_name = os.getenv(
                "AZURE_OPENAI_DEPLOYMENT_NAME"
            ) or os.getenv("OPENAI_DEPLOYMENT_NAME")

        self.aoai_api_endpoint = aoai_api_endpoint
        self.aoai_api_key = aoai_api_key
        self.aoai_api_version = aoai_api_version
        self.aoai_deployment_name = aoai_deployment_name

        try:
            if self.aoai_api_endpoint is None:
                raise ValueError("Azure OpenAI endpoint cannot be None")

            self.client = AzureOpenAI(
                azure_endpoint=self.aoai_api_endpoint,
                api_key=self.aoai_api_key,
                api_version=self.aoai_api_version,
            )
            print("=== Initialized AzureOpenAI client ===")
            print(f"AZURE_OPENAI_ENDPOINT={aoai_api_endpoint}")
            print(f"AZURE_OPENAI_API_VERSION={aoai_api_version}")
            print(f"AZURE_OPENAI_DEPLOYMENT_NAME={aoai_deployment_name}")

        except (ValueError, TypeError) as e:
            print("=== Failed to initialize AzureOpenAI client ===")
            print(e)

    def set_deployment_name(self, deployment_name: str):
        """
        Set the deployment name for Azure OpenAI
        Args:
            deployment_name: Azure OpenAI deployment name
        """
        self.aoai_deployment_name = deployment_name

    def get_aoai_client(self):
        return self.client

    def test_api_call(self):
        """
        Simple API Call
        """
        system_message = """
        You are an AI assistant that helps customers find information. As an assistant, you respond to questions in a concise and unique manner.
        You can use Markdown to answer simply and concisely, and add a personal touch with appropriate emojis.

        Add a witty joke starting with "By the way," at the end of your response. Do not mention the customer's name in the joke part.
        The joke should be related to the specific question asked.
        For example, if the question is about tents, the joke should be specifically related to tents.

        Use the given context to provide a more personalized response. Write each sentence on a new line:
        """
        context = """
            The Alpine Explorer Tent features a detachable partition to ensure privacy, 
            numerous mesh windows and adjustable vents for ventilation, and a waterproof design. 
            It also includes a built-in gear loft for storing outdoor essentials. 
            In short, it offers a harmonious blend of privacy, comfort, and convenience, making it a second home in nature!
        """
        question = "What are features of the Alpine Explorer Tent?"

        user_message = f"""
        Context: {context}
        Question: {question}
        """

        if self.aoai_deployment_name is None:
            raise ValueError("Deployment name cannot be None")

        response = self.client.chat.completions.create(
            model=self.aoai_deployment_name,
            messages=[
                {"role": "system", "content": system_message},
                {"role": "user", "content": user_message},
            ],
            temperature=0.7,
            max_tokens=300,
        )

        print(response.choices[0].message.content)



"""Question-Answer Data Generator."""

USER_AGENT = "azure-genai-utils"
USER_AGENT_HEADER_KEY = "Client-User-Agent"

try:
    import asyncio
    import nest_asyncio
    import os
    import time
    import logging
    import random
    from enum import Enum
    from functools import lru_cache
    from packaging import version
    from typing import Dict, List, Tuple, Any, Union, Optional
    from collections import defaultdict
    from azure.core import CaseInsensitiveEnumMeta
    from azure.core.tracing.decorator import distributed_trace

except ImportError as e:
    print("Import Error! Please install the required packages.")
    raise e

try:
    import pkg_resources  # type: ignore[import]

    openai_version_str = pkg_resources.get_distribution("openai").version
    openai_version = pkg_resources.parse_version(openai_version_str)
    import openai

    if openai_version >= pkg_resources.parse_version("1.0.0"):
        _RETRY_ERRORS: Tuple = (
            openai.APIConnectionError,
            openai.APIError,
            openai.APIStatusError,
        )
    else:
        _RETRY_ERRORS: Tuple = (  # type: ignore[no-redef]
            openai.APIError,  # pylint: disable=no-member
            openai.RateLimitError,  # pylint: disable=no-member
            openai.APIConnectionError,  # pylint: disable=no-member
            openai.Timeout,  # pylint: disable=no-member
        )

except ImportError as e:
    print(
        "In order to use qa, please install the 'qa_generation' extra of azure-ai-generative"
    )
    raise e

_TEMPLATES_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "prompt_templates/en"
)

logger = logging.getLogger(__name__)

_DEFAULT_AOAI_VERSION = "2024-12-01-preview"
_MAX_RETRIES = 7


def _completion_with_retries(*args, **kwargs):
    n = 1
    while True:
        try:
            if openai_version >= pkg_resources.parse_version("1.0.0"):
                if kwargs["api_type"].lower() == "azure":
                    from openai import AzureOpenAI

                    client = AzureOpenAI(
                        azure_endpoint=kwargs["api_base"],
                        api_key=kwargs["api_key"],
                        api_version=kwargs["api_version"],
                        default_headers={USER_AGENT_HEADER_KEY: USER_AGENT},
                    )
                    response = client.chat.completions.create(
                        messages=kwargs["messages"],
                        model=kwargs["deployment_id"],
                        temperature=kwargs["temperature"],
                        max_tokens=kwargs["max_tokens"],
                    )
                else:
                    from openai import OpenAI

                    client = OpenAI(
                        api_key=kwargs["api_key"],
                        default_headers={USER_AGENT_HEADER_KEY: USER_AGENT},
                    )
                    response = client.chat.completions.create(
                        messages=kwargs["messages"],
                        model=kwargs["model"],
                        temperature=kwargs["temperature"],
                        max_tokens=kwargs["max_tokens"],
                    )
                return response.choices[0].message.content, (
                    dict(response.usage) if response.usage else {}
                )
            response = openai.ChatCompletion.create(
                *args, **kwargs
            )  # pylint: disable=no-member
            return response["choices"][0].message.content, response["usage"]
        except _RETRY_ERRORS as _re:  # pylint: disable=catching-non-exception
            if n > _MAX_RETRIES:
                raise
            secs = 2**n
            msg = f"Retrying after {secs}s. API call failed due to {_re.__class__.__name__}: {_re}"
            logger.warning(msg)
            time.sleep(secs)
            n += 1
            continue


async def _completion_with_retries_async(*args, **kwargs):
    n = 1
    while True:
        try:
            if openai_version >= pkg_resources.parse_version("1.0.0"):
                if kwargs["api_type"].lower() == "azure":
                    from openai import AsyncAzureOpenAI

                    client = AsyncAzureOpenAI(
                        azure_endpoint=kwargs["api_base"],
                        api_key=kwargs["api_key"],
                        api_version=kwargs["api_version"],
                        default_headers={USER_AGENT_HEADER_KEY: USER_AGENT},
                    )
                    response = await client.chat.completions.create(
                        messages=kwargs["messages"],
                        model=kwargs["deployment_id"],
                        temperature=kwargs["temperature"],
                        max_tokens=kwargs["max_tokens"],
                    )
                else:
                    from openai import AsyncOpenAI

                    client = AsyncOpenAI(
                        api_key=kwargs["api_key"],
                        default_headers={USER_AGENT_HEADER_KEY: USER_AGENT},
                    )
                    response = await client.chat.completions.create(
                        messages=kwargs["messages"],
                        model=kwargs["model"],
                        temperature=kwargs["temperature"],
                        max_tokens=kwargs["max_tokens"],
                    )
                return response.choices[0].message.content, (
                    dict(response.usage) if response.usage else {}
                )
            response = openai.ChatCompletion.create(
                *args, **kwargs
            )  # pylint: disable=no-member
            return response["choices"][0].message.content, response["usage"]
        except _RETRY_ERRORS as _re:  # pylint: disable=catching-non-exception
            if n > _MAX_RETRIES:
                raise
            secs = 2**n
            logger.warning(
                "Retrying after %ss. API call failed due to %s: %s",
                secs,
                _re.__class__.__name__,
                _re,
            )
            await asyncio.sleep(secs)
            n += 1
            continue


class OutputStructure(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """OutputStructure defines what structure the QAs should be written to file in."""

    PROMPTFLOW = "PROMPTFLOW"
    """Chat history will be in format used by promptflow"""
    CHAT_PROTOCOL = "CHAT_PROTOCOL"
    """QAs will be in OpenAI message format"""


class QAType(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """QAType defines different types of QAs that can be generated."""

    SHORT_ANSWER = "SHORT_ANSWER"
    """
        Short answer QAs have answers that are only a few words long.
        These words are generally relevant details from text like dates, names, statistics, etc.
    """
    LONG_ANSWER = "LONG_ANSWER"
    """
        Long answer QAs have answers that are one or more sentences long.
        ex. Questions where answer is a definition: What is a {topic_from_text}?
    """
    BOOLEAN = "BOOLEAN"
    """Boolean QAs have answers that are either True or False."""
    SUMMARY = "SUMMARY"
    """
        Summary QAs have questions that ask to write a summary for text's title in a limited number of words.
        It generates just one QA.
    """
    CONVERSATION = "CONVERSATION"
    """
        Conversation QAs have questions that might reference words or ideas from previous QAs.
        ex. If previous conversation was about some topicX from text, next question might reference it
        without using its name: How does *it* compare to topicY?
    """


class QADataGenerator:
    """Class for generating Question-Answer data from text."""

    _PARSING_ERR_UNEQUAL_QA = "Parsing error: Unequal question answer count"
    _PARSING_ERR_UNEQUAL_Q_AFTER_MOD = (
        "Parsing error: Unequal question count after modification"
    )
    _PARSING_ERR_FIRST_LINE = "Parsing error: First line must be a question"

    def __init__(self, model_config: Dict):
        """Initialize QADataGenerator using Azure OpenAI details."""

        api_key = "OPENAI_API_KEY"
        api_base = "OPENAI_API_BASE"
        if version.parse(openai.version.VERSION) >= version.parse("1.0.0"):
            api_key = "AZURE_OPENAI_KEY"
            api_base = "AZURE_OPENAI_ENDPOINT"
        self._chat_completion_params = dict(
            # AOAI connection params
            api_type=(
                model_config["api_type"]
                if "api_type" in model_config
                else os.getenv("OPENAI_API_TYPE", "azure")
            ),
            api_version=(
                model_config["api_version"]
                if "api_version" in model_config
                else os.getenv("OPENAI_API_VERSION", _DEFAULT_AOAI_VERSION)
            ),
            api_base=(
                model_config["api_base"]
                if "api_base" in model_config
                else os.getenv(api_base)
            ),
            api_key=(
                model_config["api_key"]
                if "api_key" in model_config
                else os.getenv(api_key)
            ),
            # AOAI model params
            deployment_id=model_config["deployment"],
            model=model_config["model"],
            max_tokens=model_config.get("max_tokens", 2000),
            temperature=0.0,  # don't need creativity
        )

    def _validate(self, qa_type: QAType, num_questions: Optional[int]):
        if qa_type == QAType.SUMMARY and num_questions is not None:
            raise ValueError("num_questions unsupported for Summary QAType")
        if qa_type != QAType.SUMMARY and num_questions <= 0:  # type: ignore[operator]
            raise ValueError("num_questions must be an integer greater than zero")

    def _get_messages_for_qa_type(
        self, qa_type: QAType, text: str, num_questions: int
    ) -> List:
        logger.debug("Getting prompt messages for %s QA type", qa_type)
        template_filename = {
            QAType.SHORT_ANSWER: "prompt_qa_short_answer.txt",
            QAType.LONG_ANSWER: "prompt_qa_long_answer.txt",
            QAType.BOOLEAN: "prompt_qa_boolean.txt",
            QAType.SUMMARY: "prompt_qa_summary.txt",
            QAType.CONVERSATION: "prompt_qa_conversation.txt",
        }
        filename = template_filename[qa_type]
        messages = self._get_messages_from_file(filename)
        input_variables: Dict[str, Any] = {"text": text}
        if qa_type == QAType.SUMMARY:
            input_variables["num_words"] = 100
        else:
            input_variables["num_questions"] = num_questions
        messages[-1]["content"] = messages[-1]["content"].format(**input_variables)
        return messages

    def _get_messages_for_modify_conversation(self, questions: List[str]) -> List:
        messages = self._get_messages_from_file("prompt_qa_conversation_modify.txt")
        questions_str = "\n".join([f"[Q]: {q}" for q in questions])
        messages[-1]["content"] = messages[-1]["content"].format(
            questions=questions_str
        )
        return messages

    def _get_messages_from_file(self, filename: str) -> List:
        template = self._get_template(filename)
        content_list = [content.strip() for content in template.split("<|separator|>")]
        messages = [
            {"role": "system", "content": content_list[0]},  # system instructions
            {"role": "user", "content": content_list[1]},  # few-shot input
            {"role": "assistant", "content": content_list[2]},  # few-shot output
            {"role": "user", "content": content_list[3]},  # input template
        ]
        return messages

    @lru_cache
    def _get_template(self, filename) -> str:
        logger.debug("Getting prompt template from %s file", filename)
        filepath = os.path.join(_TEMPLATES_DIR, filename)
        with open(filepath, encoding="utf-8") as f:
            template = f.read()
        return template

    def _parse_qa_from_response(
        self, response_text: str
    ) -> Tuple[List[str], List[str]]:
        q_prefix, a_prefix = "[Q]: ", "[A]: "
        last_updated = None
        questions, answers = [], []
        for line in response_text.split("\n"):
            if line.startswith(q_prefix):
                questions.append(line[len(q_prefix) :])
                last_updated = "Q"
            elif line.startswith(a_prefix):
                answers.append(line[len(a_prefix) :])
                last_updated = "A"
            else:  # Q or A spread across multiple lines
                assert last_updated is not None, self._PARSING_ERR_FIRST_LINE
                if last_updated == "Q":
                    questions[-1] += "\n" + line
                else:
                    answers[-1] += "\n" + line
        return questions, answers

    def _merge_token_usage(self, token_usage: Dict, token_usage2: Dict) -> Dict:
        return {name: count + token_usage[name] for name, count in token_usage2.items()}

    def _modify_conversation_questions(self, questions) -> Tuple[List[str], Dict]:
        content, usage = _completion_with_retries(
            messages=self._get_messages_for_modify_conversation(questions),
            **self._chat_completion_params,
        )

        modified_questions, _ = self._parse_qa_from_response(content)
        # Keep proper nouns in first question of conversation
        modified_questions[0] = questions[0]
        assert len(modified_questions) == len(
            questions
        ), self._PARSING_ERR_UNEQUAL_Q_AFTER_MOD
        return modified_questions, usage

    @distributed_trace
    def export_to_file(
        self,
        output_path: str,
        qa_type: QAType,
        results: Union[List, List[List]],
        output_format: OutputStructure = OutputStructure.PROMPTFLOW,
        field_mapping: Optional[Dict[str, str]] = None,
    ):
        """
        Writes results from QA gen to a jsonl file for Promptflow batch run results is either a list of questions
        and answers or list of list of questions and answers grouped by their chunk e.g. [("How are you?",
        "I am good.")] or [ [("How are you?", "I am good.")], [("What can I do?", "Tell me a joke.")].

        :param output_path: The path to the output file.
        :type output_path: str
        :param qa_type: The type of QA data.
        :type qa_type: QAType
        :param results: The results of the QA generation.
        :type results: Union[List, List[List]]
        :param output_format: The output structure format.
        :type output_format: OutputStructure, optional
        :param field_mapping: The field mapping for the output structure.
        :type field_mapping: Optional[Dict[str, str]], optional
        """
        data_dict = defaultdict(list)

        if field_mapping is None:
            field_mapping = {
                "chat_history_key": "chat_history",
                "question_key": "question",
            }

        if not isinstance(results[0], List):
            results = [results]

        if output_format == OutputStructure.PROMPTFLOW:

            if qa_type == QAType.CONVERSATION and not (
                "chat_history_key" in field_mapping and "question_key" in field_mapping
            ):
                keys = "chat_history_key, question_key"
                raise Exception(
                    "Field mapping for Promptflow output with Conversation must contain following keys: "
                    + keys
                )
            # Only the question key is required in non-conversation cases,
            # we can default to chat_history as chat_history_key
            if "question_key" not in field_mapping:
                raise Exception(
                    f"Field mapping for Promptflow output with {qa_type} must contain following keys: question_key"
                )

            question_key = field_mapping["question_key"]
            # Set this here for parity with eval flows
            answer_key = "ground_truth"
            chat_history_key = field_mapping.get("chat_history_key", "chat_history")
            for qs_and_as in results:
                chat_history: List = []
                for question, answer in qs_and_as:
                    data_dict[chat_history_key].append(list(chat_history))
                    if qa_type == QAType.CONVERSATION:
                        # Chat History columns:
                        data_dict[question_key].append(question)
                        chat_history.append(
                            {
                                "inputs": {question_key: question},
                                "outputs": {answer_key: answer},
                            }
                        )
                    else:
                        # QnA columns:
                        data_dict[question_key].append(question)

                    data_dict[answer_key].append(
                        answer
                    )  # Consider generated answer as the ground truth
        else:
            for qs_and_as in results:
                chat_history = []
                for question, answer in qs_and_as:
                    if qa_type == QAType.CONVERSATION:
                        print(f"Chat data dict: {data_dict['messages']}\n\n")
                        chat_history.append({"role": "user", "content": question})
                        chat_history.append({"role": "assistant", "content": answer})
                        data_dict["messages"].append(list(chat_history))
                    else:
                        messages = []
                        messages.append({"role": "user", "content": question})
                        messages.append({"role": "assistant", "content": answer})
                        data_dict["messages"].append(list(messages))
        # export to jsonl file
        try:
            import pandas as pd
        except ImportError as ie:
            print("In order to write qa data to file, please install pandas")
            raise ie

        data_df = pd.DataFrame(data_dict, columns=list(data_dict.keys()))
        data_df.to_json(output_path, lines=True, orient="records")

    @distributed_trace
    def generate(
        self, text: str, qa_type: QAType, num_questions: Optional[int] = None
    ) -> Dict:
        self._validate(qa_type, num_questions)
        validated_num_questions: int = num_questions  # type: ignore[assignment]
        content, token_usage = _completion_with_retries(
            messages=self._get_messages_for_qa_type(
                qa_type, text, validated_num_questions
            ),
            **self._chat_completion_params,
        )
        questions, answers = self._parse_qa_from_response(content)
        assert len(questions) == len(answers), self._PARSING_ERR_UNEQUAL_QA
        if qa_type == QAType.CONVERSATION:
            questions, token_usage2 = self._modify_conversation_questions(questions)
            token_usage = self._merge_token_usage(token_usage, token_usage2)
        return {
            "question_answers": list(zip(questions, answers)),
            "token_usage": token_usage,
        }

    async def _modify_conversation_questions_async(
        self, questions
    ) -> Tuple[List[str], Dict]:
        content, usage = await _completion_with_retries_async(
            messages=self._get_messages_for_modify_conversation(questions),
            **self._chat_completion_params,
        )

        modified_questions, _ = self._parse_qa_from_response(content)
        # Keep proper nouns in first question of conversation
        modified_questions[0] = questions[0]
        assert len(modified_questions) == len(
            questions
        ), self._PARSING_ERR_UNEQUAL_Q_AFTER_MOD
        return modified_questions, usage

    @distributed_trace
    async def generate_async(
        self, text: str, qa_type: QAType, num_questions: Optional[int] = None
    ) -> Dict:
        self._validate(qa_type, num_questions)
        validated_num_questions: int = num_questions  # type: ignore[assignment]
        content, token_usage = await _completion_with_retries_async(
            messages=self._get_messages_for_qa_type(
                qa_type, text, validated_num_questions
            ),
            **self._chat_completion_params,
        )
        questions, answers = self._parse_qa_from_response(content)
        assert len(questions) == len(answers), self._PARSING_ERR_UNEQUAL_QA
        if qa_type == QAType.CONVERSATION:
            questions, token_usage2 = await self._modify_conversation_questions_async(
                questions
            )
            token_usage = self._merge_token_usage(token_usage, token_usage2)
        return {
            "question_answers": list(zip(questions, answers)),
            "token_usage": token_usage,
        }


class CustomQADataGenerator(QADataGenerator):
    """
    Create a custom QA data generator by providing the path to the templates directory.
    """

    def __init__(self, templates_dir: str, **kwargs):
        self.templates_dir = templates_dir
        super().__init__(**kwargs)

    def _get_template(self, filename) -> str:
        logger.debug("Getting prompt template from %s file", filename)
        filepath = os.path.join(self.templates_dir, filename)
        with open(filepath, encoding="utf-8") as f:
            template = f.read()
        return template

    def _get_messages_for_qa_type(
        self, qa_type: QAType, text: str, num_questions: int
    ) -> List:
        logger.debug("Getting prompt messages for %s QA type", qa_type)
        template_filename = {
            QAType.SHORT_ANSWER: "prompt_qa_short_answer.txt",
            QAType.LONG_ANSWER: "prompt_qa_long_answer.txt",
            QAType.BOOLEAN: "prompt_qa_boolean.txt",
            QAType.SUMMARY: "prompt_qa_summary.txt",
            QAType.CONVERSATION: "prompt_qa_conversation.txt",
        }
        filename = template_filename[qa_type]
        messages = self._get_messages_from_file(filename)
        input_variables: Dict[str, Any] = {"text": text}
        if qa_type == QAType.SUMMARY:
            input_variables["num_words"] = 100
        else:
            input_variables["num_questions"] = num_questions
        messages[-1]["content"] = messages[-1]["content"].format(**input_variables)
        return messages

    def _get_messages_for_modify_conversation(self, questions: List[str]) -> List:
        messages = self._get_messages_from_file("prompt_qa_conversation_modify.txt")
        questions_str = "\n".join([f"[Q]: {q}" for q in questions])
        messages[-1]["content"] = messages[-1]["content"].format(
            questions=questions_str
        )
        return messages


def convert_to_oai_format(
    qa_pair,
    system_prompt_msg="You're an AI assistant that guides a user to the location of your CS center",
):
    """
    Convert the QA pair to the jsonl format required by the OpenAI API.

    Args:
        qa_pair: list of dictionaries or list of lists containing the QA pairs
        system_prompt_msg: message to be displayed as the system prompt

    Returns:
        formatted_data: jsonl format data for OpenAI API
    """
    if isinstance(qa_pair, list):
        formatted_data = []
        for qa in qa_pair:

            sample = [{"role": "system", "content": system_prompt_msg}]

            if isinstance(qa, list):  # multi-turn
                for qa_ in qa:
                    if isinstance(qa_, dict):
                        user_message = {"role": "user", "content": qa_["QUESTION"]}
                        assistant_message = {
                            "role": "assistant",
                            "content": qa_["ANSWER"],
                        }
                    else:
                        user_message = {"role": "user", "content": qa_[0]}
                        assistant_message = {"role": "assistant", "content": qa_[1]}
                    sample.append(user_message)
                    sample.append(assistant_message)
            else:  # single turn
                if isinstance(qa, dict):
                    user_message = {"role": "user", "content": qa["QUESTION"]}
                    assistant_message = {"role": "assistant", "content": qa["ANSWER"]}
                else:
                    user_message = {"role": "user", "content": qa[0]}
                    assistant_message = {"role": "assistant", "content": qa[1]}
                sample.append(user_message)
                sample.append(assistant_message)

            msg = {"messages": sample}
            formatted_data.append(msg)
        random.shuffle(formatted_data)
        return formatted_data

    else:
        print("Argument is not a list")
        return None


nest_asyncio.apply()


def generate_qas(
    input_texts: List[str],
    qa_generator: Union[QADataGenerator, CustomQADataGenerator],
    qa_type: QAType = QAType.LONG_ANSWER,
    num_questions: int = 3,
    concurrency: int = 3,
) -> List[List[Dict[str, str]]]:
    """
    Generates question-answer pairs for a list of input texts using Azure GenAI QA generator.

    Args:
        input_texts (List[str]): List of texts to generate QAs for.
        qa_generator (Union[QADataGenerator, CustomQADataGenerator]): QA generator instance.
        qa_type (QAType): Type of QA to generate (e.g., LONG_ANSWER, SHORT_ANSWER). Default is LONG_ANSWER.
        num_questions (int): Number of questions to generate per text. Default is 3.
        concurrency (int): Number of concurrent calls to the generator. Default is 3.

    Returns:
        List[List[Dict[str, str]]]: A list of lists containing question-answer pairs for each input text.
    """

    sem = asyncio.Semaphore(concurrency)

    async def generate_async(text: str) -> Dict:
        async with sem:
            result = await qa_generator.generate_async(
                text=text,
                qa_type=qa_type,
                num_questions=num_questions,
            )
            # print(f"Task completed for input: {text}")
            return result

    async def main():
        tasks = [generate_async(text) for text in input_texts]
        results = await asyncio.gather(*tasks, return_exceptions=True)

        for i, result in enumerate(results):
            if isinstance(result, Exception):
                print(f"Task {i} failed with error: {result}")
            else:
                print(f"Task {i} completed successfully")

        return results

    # Run the event loop in the current context
    return (
        asyncio.run(main())
        if not asyncio.get_event_loop().is_running()
        else asyncio.get_event_loop().create_task(main())
    )  # type: ignore



import importlib
from typing import Any
from .bing import BingSearch

_module_lookup = {
    "BingSearch": "tools.bing",
}


def __getattr__(name: str) -> Any:
    if name in _module_lookup:
        module = importlib.import_module(_module_lookup[name])
        return getattr(module, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


__all__ = ["BingSearch"]



from azure_genai_utils.rag.base import RetrievalChain
from langchain_community.document_loaders import (
    PDFPlumberLoader,
    PyPDFLoader,
    PyMuPDFLoader,
)
from langchain_text_splitters import RecursiveCharacterTextSplitter
from typing import List, Annotated, Optional


class PDFRetrievalChain(RetrievalChain):
    def __init__(
        self,
        source_uri: Annotated[List[str], "Source URI"],
        k: Annotated[int, "Number of results to retrieve"] = 10,
        loader_type: Annotated[str, "PDF Loader Type"] = "PDFPlumber",
        chunk_size: Annotated[int, "Chunk Size"] = 300,
        chunk_overlap: Annotated[int, "Chunk Overlap"] = 50,
        **kwargs,  # Additional keyword arguments for parent class
    ):
        """
        Initialize the retrieval chain with the source URI and loader type.

        :param source_uri: URI of the PDF source
        :param loader_type: Type of PDF loader to use. Options are "PDFPlumber", "PyPDF", or "PyMuPDF".
        """
        super().__init__(**kwargs)
        self.source_uri = source_uri  # Override if needed
        self.k = k  # Override if needed
        self.loader_type = loader_type  # Specific to PDFRetrievalChain
        self.loader_map = {
            "PDFPlumber": PDFPlumberLoader,
            "PyPDF": PyPDFLoader,
            "PyMuPDF": PyMuPDFLoader,
        }
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    def get_loader(self, loader_type: Optional[str] = None):
        """
        Returns the loader class based on the loader type.

        :param loader_type: Optional; the type of loader to use.
        :return: The corresponding loader class.
        """
        loader_type = loader_type or self.loader_type
        if loader_type not in self.loader_map:
            raise ValueError(
                f"Unsupported loader type: {loader_type}. "
                f"Supported types are: {list(self.loader_map.keys())}"
            )
        return self.loader_map[loader_type]

    def load_documents(self, source_uris: List[str]):
        """
        Load documents using the selected loader type.

        :param source_uris: List of source URIs
        :return: List of loaded documents
        """
        docs = []
        loader_class = self.get_loader()

        for source_uri in source_uris:
            loader = loader_class(source_uri)
            docs.extend(loader.load())

        return docs

    def create_text_splitter(self):
        """
        Create a text splitter for processing the documents.

        :return: RecursiveCharacterTextSplitter instance
        """
        return RecursiveCharacterTextSplitter(
            chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap
        )



import uuid
from langchain_core.messages import AIMessageChunk
from typing import Any, Dict, List, Callable, Optional
from dataclasses import dataclass
from langchain_core.agents import AgentAction, AgentStep
from langchain.agents.output_parsers.tools import ToolAgentAction
from langchain_core.messages import BaseMessage, HumanMessage, AIMessage
from langchain_core.runnables import RunnableConfig
from langgraph.graph.state import CompiledStateGraph


def random_uuid():
    return str(uuid.uuid4())


def stream_response(response, return_output=False):
    """
    Streams the response from the AI ​​model, processing each chunk as it is output.

    This function iterates over each item in the `response` iterable. If the item is an instance of `AIMessageChunk`,
    it extracts the contents of the chunk and outputs it. If the item is a string, it outputs the string directly. Optionally, the function
    can return a concatenated string of all the response chunks.

    :param response (iterable): An iterable of response chunks that can be `AIMessageChunk` objects or strings.
    :param return_output (bool, optional): If True, the function returns the concatenated response string. Default is False.
    :return str: The concatenated response string if `return_output` is True. Otherwise, nothing is returned.
    """
    answer = ""
    for token in response:
        if isinstance(token, AIMessageChunk):
            if isinstance(token.content, list):
                answer += "".join(str(item) for item in token.content)
            else:
                answer += token.content
            print(token.content, end="", flush=True)
        elif isinstance(token, str):
            answer += token
            print(token, end="", flush=True)
    if return_output:
        return answer


def tool_callback(tool) -> None:
    """
    Callback function for tool usage.
    """
    print("Toll Usage]")
    print(f"Tool: {tool.get('tool')}")
    if tool_input := tool.get("tool_input"):
        for k, v in tool_input.items():
            print(f"{k}: {v}")
    print(f"Log: {tool.get('log')}")


def observation_callback(observation) -> None:
    """
    Callback function for observation.
    """
    print("[Observation]")
    print(f"Observation: {observation.get('observation')}")


def result_callback(result: str) -> None:
    """
    Callback function for the final result.
    """
    print("[Final Result]")
    print(result)


@dataclass
class AgentCallbacks:
    """
    Data class containing agent callback functions.
    :param tool_callback: Callback function for tool usage.
    :param observation_callback: Callback function for observation.
    :param result_callback: Callback function for the final result.
    """

    tool_callback: Callable[[Dict[str, Any]], None] = tool_callback
    observation_callback: Callable[[Dict[str, Any]], None] = observation_callback
    result_callback: Callable[[str], None] = result_callback


class AgentStreamParser:
    """
    A class for parsing and handling an agent's stream output.
    """

    def __init__(self, callbacks: AgentCallbacks = AgentCallbacks()):
        self.callbacks = callbacks
        self.output = None

    def process_agent_steps(self, step: Dict[str, Any]) -> None:
        if "actions" in step:
            self._process_actions(step["actions"])
        elif "steps" in step:
            self._process_observations(step["steps"])
        elif "output" in step:
            self._process_result(step["output"])

    def _process_actions(self, actions: List[Any]) -> None:
        for action in actions:
            if isinstance(action, (AgentAction, ToolAgentAction)) and hasattr(
                action, "tool"
            ):
                self._process_tool_call(action)

    def _process_tool_call(self, action: Any) -> None:
        tool_action = {
            "tool": getattr(action, "tool", None),
            "tool_input": getattr(action, "tool_input", None),
            "log": getattr(action, "log", None),
        }
        self.callbacks.tool_callback(tool_action)

    def _process_observations(self, observations: List[Any]) -> None:
        for observation in observations:
            observation_dict = {}
            if isinstance(observation, AgentStep):
                observation_dict["observation"] = getattr(
                    observation, "observation", None
                )
            self.callbacks.observation_callback(observation_dict)

    def _process_result(self, result: str) -> None:
        self.callbacks.result_callback(result)
        self.output = result


def pretty_print_messages(messages: list[BaseMessage]):
    for message in messages:
        message.pretty_print()


depth_colors = {
    1: "\033[96m",  # Bright Cyan (First Layer)
    2: "\033[93m",  # Yellow (Second Layer)
    3: "\033[94m",  # Light Green (Third Layer)
    4: "\033[95m",  # Purple (Fourth Layer)
    5: "\033[92m",  # Light Blue (Fifth Layer)
    "default": "\033[96m",  # Default Color
    "reset": "\033[0m",  # Reset Color
}


def is_terminal_dict(data):
    """Check if dictionary is a terminal dictionary."""
    if not isinstance(data, dict):
        return False
    for value in data.values():
        if isinstance(value, (dict, list)) or hasattr(value, "__dict__"):
            return False
    return True


def format_terminal_dict(data):
    """Format terminal dictionary."""
    items = []
    for key, value in data.items():
        if isinstance(value, str):
            items.append(f'"{key}": "{value}"')
        else:
            items.append(f'"{key}": {value}')
    return "{" + ", ".join(items) + "}"


def _display_message_tree(data, indent=0, node=None, is_root=False):
    """Print JSON object tree structure without type information."""
    spacing = " " * indent * 4
    color = depth_colors.get(indent + 1, depth_colors["default"])

    if isinstance(data, dict):
        if not is_root and node is not None:
            if is_terminal_dict(data):
                print(
                    f'{spacing}{color}{node}{depth_colors["reset"]}: {format_terminal_dict(data)}'
                )
            else:
                print(f'{spacing}{color}{node}{depth_colors["reset"]}:')
                for key, value in data.items():
                    _display_message_tree(value, indent + 1, key)
        else:
            for key, value in data.items():
                _display_message_tree(value, indent + 1, key)

    elif isinstance(data, list):
        if not is_root and node is not None:
            print(f'{spacing}{color}{node}{depth_colors["reset"]}:')

        for index, item in enumerate(data):
            print(f'{spacing}    {color}index [{index}]{depth_colors["reset"]}')
            _display_message_tree(item, indent + 1)

    elif hasattr(data, "__dict__") and not is_root:
        if node is not None:
            print(f'{spacing}{color}{node}{depth_colors["reset"]}:')
        _display_message_tree(data.__dict__, indent)

    else:
        if node is not None:
            if isinstance(data, str):
                value_str = f'"{data}"'
            else:
                value_str = str(data)

            print(f'{spacing}{color}{node}{depth_colors["reset"]}: {value_str}')


def display_message_tree(message):
    """Message tree display main function."""
    if isinstance(message, BaseMessage):
        _display_message_tree(message.__dict__, is_root=True)
    else:
        _display_message_tree(message, is_root=True)


class ToolChunkHandler:
    """Class for handling and managing Tool Message chunks."""

    def __init__(self):
        self._reset_state()

    def _reset_state(self) -> None:
        self.gathered = None
        self.first = True
        self.current_node = None
        self.current_namespace = None

    def _should_reset(self, node: str | None, namespace: str | None) -> bool:
        """Check state reset condition"""
        # Default behavior is to not reset if both parameters are None
        if node is None and namespace is None:
            return False

        # If only node is set
        if node is not None and namespace is None:
            return self.current_node != node

        # If only namespace is set
        if namespace is not None and node is None:
            return self.current_namespace != namespace

        # If both are set
        return self.current_node != node or self.current_namespace != namespace

    def process_message(
        self,
        chunk: AIMessageChunk,
        node: str | None = None,
        namespace: str | None = None,
    ) -> None:
        """Process message chunk"""
        if self._should_reset(node, namespace):
            self._reset_state()

        self.current_node = node if node is not None else self.current_node
        self.current_namespace = (
            namespace if namespace is not None else self.current_namespace
        )

        self._accumulate_chunk(chunk)
        return self._display_tool_calls()

    def _accumulate_chunk(self, chunk: AIMessageChunk) -> None:
        """Accumulate chunk content"""
        self.gathered = chunk if self.first else self.gathered + chunk
        self.first = False

    def _display_tool_calls(self) -> None:
        """Display tool calls"""
        if (
            self.gathered
            and not self.gathered.content
            and self.gathered.tool_call_chunks
            and self.gathered.tool_calls
        ):
            return self.gathered.tool_calls[0]["args"]


def get_role_from_messages(msg):
    if isinstance(msg, HumanMessage):
        return "user"
    elif isinstance(msg, AIMessage):
        return "assistant"
    else:
        return "assistant"


def messages_to_history(messages):
    return "\n".join(
        [f"{get_role_from_messages(msg)}: {msg.content}" for msg in messages]
    )


def stream_graph(
    graph: CompiledStateGraph,
    inputs: dict,
    config: RunnableConfig,
    node_names: List[str] = [],
    callback: Optional[Callable] = None,
):
    """
    Stream the output of the LangGraph execution by printing it.

    param: graph (CompiledStateGraph): The compiled LangGraph object to execute
    param: inputs (dict): The input dictionary to pass to the graph
    param: config (RunnableConfig): The execution configuration
    param: node_names (List[str], optional): The list of node names to output. Default is an empty list
    param: callback (Callable, optional): The callback function for each chunk processing. Default is None
    return: None
    """
    prev_node = ""
    for chunk_msg, metadata in graph.stream(inputs, config, stream_mode="messages"):
        curr_node = metadata["langgraph_node"]

        # Process only if node_names is empty or the current node is in node_names
        if not node_names or curr_node in node_names:
            if callback:
                callback({"node": curr_node, "content": chunk_msg.content})
            else:
                # Print the node name only if it has changed
                if curr_node != prev_node:
                    print("\n" + "=" * 50)
                    print(f"🔄 Node: \033[1;36m{curr_node}\033[0m 🔄")
                    print("- " * 25)
                print(chunk_msg.content, end="", flush=True)

            prev_node = curr_node


def invoke_graph(
    graph: CompiledStateGraph,
    inputs: dict,
    config: RunnableConfig,
    node_names: List[str] = [],
    callback: Optional[Callable] = None,
):
    """
    Pretty-prints the output of the LangGraph execution by streaming it.

    param: graph (CompiledStateGraph): The compiled LangGraph object to execute
    param: inputs (dict): The input dictionary to pass to the graph
    param: config (RunnableConfig): The execution configuration
    param: node_names (List[str], optional): The list of node names to output. Default is an empty list
    param: callback (Callable, optional): The callback function for each chunk processing. Default is None
    return: None
    """

    def format_namespace(namespace):
        return namespace[-1].split(":")[0] if len(namespace) > 0 else "root graph"

    for namespace, chunk in graph.stream(
        inputs, config, stream_mode="updates", subgraphs=True
    ):
        for node_name, node_chunk in chunk.items():
            # Filter by node_names if it is not empty
            if len(node_names) > 0 and node_name not in node_names:
                continue

            if callback is not None:
                callback({"node": node_name, "content": node_chunk})
            else:
                print("\n" + "=" * 50)
                formatted_namespace = format_namespace(namespace)
                if formatted_namespace == "root graph":
                    print(f"🔄 Node: \033[1;36m{node_name}\033[0m 🔄")
                else:
                    print(
                        f"🔄 Node: \033[1;36m{node_name}\033[0m in [\033[1;33m{formatted_namespace}\033[0m] 🔄"
                    )
                print("- " * 25)

                # Print the chunk data of the node
                if isinstance(node_chunk, dict):
                    for k, v in node_chunk.items():
                        if isinstance(v, BaseMessage):
                            v.pretty_print()
                        elif isinstance(v, list):
                            for list_item in v:
                                if isinstance(list_item, BaseMessage):
                                    list_item.pretty_print()
                                else:
                                    print(list_item)
                        elif isinstance(v, dict):
                            for node_chunk_key, node_chunk_value in node_chunk.items():
                                print(f"{node_chunk_key}:\n{node_chunk_value}")
                        else:
                            print(f"\033[1;32m{k}\033[0m:\n{v}")
                else:
                    if node_chunk is not None:
                        for item in node_chunk:
                            print(item)
                print("=" * 50)



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)  # Set this to the lowest level you want to capture

# Create console handler with a higher log level
console_handler = logging.StreamHandler()
console_handler.setLevel(
    logging.DEBUG
)  # Set this to the lowest level you want to capture

# Create file handler which logs even debug messages
file_handler = logging.FileHandler("debug.log")
file_handler.setLevel(logging.DEBUG)  # Set this to the lowest level you want to capture

# Create formatter and add it to the handlers
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
console_handler.setFormatter(formatter)
file_handler.setFormatter(formatter)

# Add the handlers to the logger
logger.addHandler(console_handler)
logger.addHandler(file_handler)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()
setup(
    name="azure-genai-utils",
    version="0.0.2.14",
    description="Utility functions for Azure GenAI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    url="https://github.com/daekeun-ml/azure-genai-utils",
    install_requires=[
        "langchain>=0.3.0",
        "langgraph>=0.2.62",
        "audiomentations>=0.38.0",
        "azure-identity",
        "azure-ai-documentintelligence==1.0.0",
        "azure-search-documents==11.5.2",
        "azure-cognitiveservices-speech>=1.42.0",
        "azure-ai-contentsafety==1.0.0",
        "azure-ai-inference==1.0.0b7",
        "tiktoken~=0.8.0",
        "openai>=1.59.7",
        "python-dotenv==1.0.1",
        "pybase64",
    ],
    packages=find_packages(exclude=[]),
    keywords=[
        "azure",
        "genai",
        "azure-genai-utils",
        "azure-genai",
        "langchain",
        "langgraph",
    ],
    python_requires=">=3.8",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
)



import time
import logging
from typing import Optional
import tiktoken


class Tokenizer:
    """Handles LLM tokenizing using the tiktoken library."""

    def __init__(self, model_name: str, logger: Optional[logging.Logger] = None):
        self.model_name = model_name
        self.logger = logger

        # Get fast tokenizer for model_name
        # NOTE: will look for models with alike prefixes if not found directly
        self.set_encoding(model_name)

    def count_tokens(self, input: str) -> int:
        # Count tokens, including special tokens like <|endofprompt|>
        return len(self.encoding.encode(input, allowed_special="all"))

    def set_encoding(self, model_name: str) -> None:
        # See: tiktoken mapping of model names here:
        #  https://github.com/openai/tiktoken/blob/main/tiktoken/model.py#L12

        start = time.time()

        try:
            encoding = tiktoken.encoding_for_model(model_name)
        except KeyError:
            self._log(
                f"Couldn't find encoding for '{model_name}'", log_level=logging.WARNING
            )

            # if chat model, return chat encoding
            if "chat" in model_name or "gpt-3.5" in model_name:
                encoding = tiktoken.get_encoding("cl100k_base")

            else:
                # Default to encoding for text & codex models
                encoding = tiktoken.get_encoding("p50k_base")

        end = time.time()

        self._log(
            f"Encoder set to '{encoding.name}'. "
            + f"Took {(end - start) * 1e3:.2f}ms to load."
        )

        self.encoding = encoding

    def _log(self, message: str, log_level: int = logging.INFO):
        if self.logger:
            self.logger.log(log_level, message)
